;; Generated package description from consult-hoogle.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "consult-hoogle" "0.5.0" "Hoogle frontend using consult" '((emacs "27.1") (consult "2.0")) :commit "384959016022d071464dc6e611e4fcded562834e" :authors '(("rahguzar" . "rahguzar@mailbox.org")) :maintainer '("rahguzar" . "rahguzar@mailbox.org") :keywords '("docs" "languages") :url "https://codeberg.org/rahguzar/consult-hoogle")
